/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gst/gstelement.h>


/* Element signals and args */
enum {
  STATE_CHANGE,
  NEW_PAD,
  NEW_GHOST_PAD,
  ERROR,
  LAST_SIGNAL
};

enum {
  ARG_0,
  /* FILL ME */
};


static void gst_element_class_init(GstElementClass *klass);
static void gst_element_init(GstElement *element);


static GstElement *parent_class = NULL;
static guint gst_element_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_element_get_type(void) {
  static GtkType element_type = 0;

  if (!element_type) {
    static const GtkTypeInfo element_info = {
      "GstElement",
      sizeof(GstElement),
      sizeof(GstElementClass),
      (GtkClassInitFunc)gst_element_class_init,
      (GtkObjectInitFunc)gst_element_init,
      (GtkArgSetFunc)NULL,
      (GtkArgGetFunc)NULL,
      (GtkClassInitFunc)NULL,
    };
    element_type = gtk_type_unique(GST_TYPE_OBJECT,&element_info);
  }
  return element_type;
}

static void
gst_element_class_init(GstElementClass *klass) {
  GtkObjectClass *gtkobject_class;

  gtkobject_class = (GtkObjectClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_OBJECT);

  gst_element_signals[STATE_CHANGE] =
    gtk_signal_new("state_change",GTK_RUN_LAST,gtkobject_class->type,
                   GTK_SIGNAL_OFFSET(GstElementClass,state_change),
                   gtk_marshal_NONE__INT,GTK_TYPE_NONE,1,
                   GTK_TYPE_INT);
  gst_element_signals[NEW_PAD] =
    gtk_signal_new("new_pad",GTK_RUN_LAST,gtkobject_class->type,
                   GTK_SIGNAL_OFFSET(GstElementClass,new_pad),
                   gtk_marshal_NONE__POINTER,GTK_TYPE_NONE,1,
                   GTK_TYPE_POINTER);
  gst_element_signals[NEW_GHOST_PAD] =
    gtk_signal_new("new_ghost_pad",GTK_RUN_LAST,gtkobject_class->type,
                   GTK_SIGNAL_OFFSET(GstElementClass,new_ghost_pad),
                   gtk_marshal_NONE__POINTER,GTK_TYPE_NONE,1,
                   GTK_TYPE_POINTER);
  gst_element_signals[ERROR] =
    gtk_signal_new("error",GTK_RUN_LAST,gtkobject_class->type,
                   GTK_SIGNAL_OFFSET(GstElementClass,error),
                   gtk_marshal_NONE__POINTER,GTK_TYPE_NONE,1,
                   GTK_TYPE_POINTER);


  gtk_object_class_add_signals(gtkobject_class,gst_element_signals,LAST_SIGNAL);

  klass->start = NULL;
  klass->stop = NULL;
  klass->change_state = gst_element_change_state;
}

static void gst_element_init(GstElement *element) {
}

/**
 * gst_element_new:
 *
 * Create a new element.
 *
 * Returns: new element
 */
GstElement *gst_element_new() {
  return GST_ELEMENT(gtk_type_new(GST_TYPE_ELEMENT));
}

/**
 * gst_element_add_pad:
 * @element: element to add pad to
 * @pad: pad to add
 *
 * Add a pad (connection point) to the element, setting the parent of the
 * pad to the element (and thus adding a reference).
 */
void gst_element_add_pad(GstElement *element,GstPad *pad) {
  g_return_if_fail(element != NULL);
  g_return_if_fail(GST_IS_ELEMENT(element));
  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));

  /* set the pad's parent */
  gst_pad_set_parent(pad,GST_OBJECT(element));

  /* add it to the list */
  element->pads = g_list_append(element->pads,pad);
  element->numpads++;

  /* emit the NEW_PAD signal */
  gtk_signal_emit(GTK_OBJECT(element),gst_element_signals[NEW_PAD],pad);
}

/**
 * gst_element_add_ghost_pad:
 * @element: element to add ghost pad to
 * @pad: ghost pad to add
 *
 * Add a ghost pad to the element, setting the ghost parent of the pad to
 * the element (and thus adding a reference).
 */
void gst_element_add_ghost_pad(GstElement *element,GstPad *pad) {
  g_return_if_fail(element != NULL);
  g_return_if_fail(GST_IS_ELEMENT(element));
  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));

  /* set the pad's parent */
  gst_pad_add_ghost_parent(pad,GST_OBJECT(element));

  /* add it to the list */
  element->pads = g_list_append(element->pads,pad);
  element->numpads++;

  /* emit the NEW_PAD signal */
  gtk_signal_emit(GTK_OBJECT(element),gst_element_signals[NEW_GHOST_PAD],pad);
}

/**
 * gst_element_get_pad:
 * @element: element to find pad of
 * @name: name of pad to retrieve
 *
 * Retrieve a pad from the element by name.
 *
 * Returns: requested pad if found, otherwise NULL.
 */
GstPad *gst_element_get_pad(GstElement *element,gchar *name) {
  GList *walk;

  g_return_if_fail(element != NULL);
  g_return_if_fail(GST_IS_ELEMENT(element));
  g_return_if_fail(name != NULL);
  g_return_if_fail(element->numpads != 0);

  walk = element->pads;
  while (walk) {
    if (!strcmp(((GstPad *)(walk->data))->name,name))
      return (GstPad *)(walk->data);
    walk = g_list_next(walk);
  }
}

/**
 * gst_element_get_pad_list:
 * @element: element to get pads of
 *
 * Retrieve a list of the pads associated with the element.
 *
 * Returns: <type>GList</type> of pads
 */
GList *gst_element_get_pad_list(GstElement *element) {
  g_return_if_fail(element != NULL);
  g_return_if_fail(GST_IS_ELEMENT(element));

  return element->pads;
}

/**
 * gst_element_connect:
 * @src: element containing source pad
 * @srcpadname: name of pad in source element
 * @dest: element containing destination pad
 * @destpadname: name of pad in destination element
 *
 * Connect the two named pads of the source and destination elements.
 * Side effect is that if one of the pads has no parent, it becomes a
 * child of the parent of the other element.  If they have different
 * parents, the connection fails.
 */
void gst_element_connect(GstElement *src,gchar *srcpadname,
                         GstElement *dest,gchar *destpadname) {
  GstPad *srcpad,*destpad;
  GstObject *srcparent,*destparent;

  g_return_if_fail(src != NULL);
  g_return_if_fail(GST_IS_ELEMENT(src));
  g_return_if_fail(srcpadname != NULL);
  g_return_if_fail(dest != NULL);
  g_return_if_fail(GST_IS_ELEMENT(dest));
  g_return_if_fail(destpadname != NULL);

  srcpad = gst_element_get_pad(src,srcpadname);
  destpad = gst_element_get_pad(dest,destpadname);

  g_return_if_fail(srcpad != NULL);
  g_return_if_fail(destpad != NULL);

  srcparent = gst_object_get_parent(GST_OBJECT(src));
  destparent = gst_object_get_parent(GST_OBJECT(dest));
  /* we can't do anything if neither have parents */
  if ((srcparent == NULL) && (destparent == NULL))
    return;

  /* and we have to make sure that they have the same parents... */
  if ((srcparent == NULL) && (destparent == NULL)) {
    if (srcparent != destparent)
      return;
  }
}

void gst_element_error(GstElement *element,gchar *error) {
  gchar *name = gst_object_get_name(GST_OBJECT(element));
  g_print("error in element '%s': %s\n",name,error);

  gtk_signal_emit(GTK_OBJECT(element),gst_element_signals[ERROR],error);
}


/**
 * gst_element_set_state:
 * @element: element to change state of
 * @state: new element state
 *
 * Sets the state of the element, but more importantly fires off a signal
 * indicating the new state.  You can clear state by simply prefixing the
 * GstElementState value with ~, it will be detected and used to turn off
 * that bit.
 *
 * Returns: whether or not the state was successfully set.
 */
gboolean gst_element_set_state(GstElement *element,GstElementState state) {
  GstElementClass *oclass;
  gboolean stateset = FALSE;

//  g_print("gst_element_set_state(\"%s\",%08lx)\n",
//          gst_object_get_name(GST_OBJECT(element)),state);

  g_return_if_fail(element != NULL);
  g_return_if_fail(GST_IS_ELEMENT(element));

  oclass = GST_ELEMENT_CLASS(GTK_OBJECT(element)->klass);

  if (oclass->change_state)
    stateset = (oclass->change_state)(element,state);

  /* if a state *set* failed, unset it immediately */
/*
  if (!(state & GST_STATE_MAX) && !stateset) {
    g_print("set state failed miserably, forcing unset\n");
    if (oclass->change_state)
      stateset = (oclass->change_state)(element,~state);
    return FALSE;
  }*/

  return stateset;
}

/* class function to set the state of a simple element */
gboolean gst_element_change_state(GstElement *element,
                                  GstElementState state) {
  g_return_if_fail(element != NULL);
  g_return_if_fail(GST_IS_ELEMENT(element));

//  g_print("gst_element_change_state(\"%s\",%d)\n",
//          gst_object_get_name(GST_OBJECT(element)),state);

  /* deal with the inverted state */
//  g_print("changing element state, was %08lx",GST_FLAGS(element));
  if (state & GST_STATE_MAX)
    GST_FLAG_UNSET(element,~state);
  else
    GST_FLAG_SET(element,state);
//  g_print(", is now %08lx\n",GST_FLAGS(element));
  gtk_signal_emit(GTK_OBJECT(element),gst_element_signals[STATE_CHANGE],
                  state);
}
